package com.citrix.mvpn

import android.app.Activity
import android.content.Context
import android.os.Handler
import android.os.Message
import android.os.Messenger
import androidx.annotation.VisibleForTesting
import com.citrix.Log
import com.citrix.mvpn.TunnelUtility.isMainThread
import com.citrix.mvpn.api.MicroVPNSDK
import io.reactivex.Completable
import io.reactivex.android.schedulers.AndroidSchedulers
import java.lang.ref.WeakReference

class TunnelInitializer(private val mContext: Context, callback: Callback?) {

    enum class TunnelState {
        NOT_REQUIRED, STARTED, TUNNELLED, FAILED
    }

    @set:VisibleForTesting
    var tunnelState = TunnelState.NOT_REQUIRED
    private val mReference: WeakReference<Callback?> = WeakReference(callback)

    fun tunnel() {
        tunnelState = TunnelState.STARTED
        if (MicroVPNSDK.isNetworkTunnelRunning(mContext)) {
            Log.d(TunnelConstants.TAG, "TunnelInitializer.tunnel(), Tunnel is already established")
            tunnelState = TunnelState.TUNNELLED
            val callback = mReference.get()
            callback?.onTunnelEstablished()
        } else {
            Log.d(TunnelConstants.TAG, "TunnelInitializer.tunnel(), Establishing tunnel")
            Completable.fromAction { tunnelInternal() }
                .subscribeOn(AndroidSchedulers.mainThread())
                .subscribe()
        }
    }

    /*
     * This function should be called from Main thread
     */
    @VisibleForTesting
    fun tunnelInternal() {
        check(isMainThread) { "tunnelInternal() should be called from the Main thread" }
        val handler: Handler = TunnelHandler(Handler.Callback { msg: Message ->
                Log.i(TunnelConstants.TAG, "In internal handle message")
                val value = msg.what
                Log.i(TunnelConstants.TAG, "Value obtained $value")
                val callback = mReference.get()
                if (callback != null) {
                    when (value) {
                        TunnelHandler.SUCCESS -> {
                            tunnelState = TunnelState.TUNNELLED
                            callback.onTunnelEstablished()
                        }
                        TunnelHandler.SESSION_EXPIRED -> {
                            tunnelState = TunnelState.FAILED
                            callback.onSessionExpired()
                        }
                        else -> {
                            tunnelState = TunnelState.FAILED
                            callback.onFailure()
                        }
                    }
                }
                true
            })
        startTunnel(mContext, Messenger(handler))
    }

    fun canProceed(): Boolean {
        val canLoadUrl = tunnelState == TunnelState.TUNNELLED ||
                tunnelState == TunnelState.NOT_REQUIRED ||
                tunnelState == TunnelState.FAILED
        Log.d(TunnelConstants.TAG, "TunnelInitializer.canProceed(), canLoadUrl : $canLoadUrl, tunnelState : $tunnelState")
        return canLoadUrl
    }

    private fun startTunnel(context: Context, messenger: Messenger) {
        Log.i(TunnelConstants.TAG, "Starting tunnel")
        if (context is Activity) {
            MicroVPNSDK.startTunnel(context as? Activity, messenger)
        } else {
            MicroVPNSDK.startTunnel(context, messenger)
        }
    }

    interface Callback {
        fun onTunnelEstablished()
        fun onFailure()
        fun onSessionExpired()
    }
}